#!/usr/bin/env python3
"""
Contains stuff around the internal image storage mechanism
of gimp files.

Generally speaking, the user should not care about anything
in this file.
"""
import zlib
import PIL.Image
from .gimpIOBase import IO, GimpIOBase


class GimpChannel(GimpIOBase):
	"""
	Represents a single channel or mask in a gimp image
	"""
	def __init__(self, parent, name='', image=None):
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self.name = name
		self._imageHierarchy = None
		self._imageHierarchyPtr = None
		if image is not None: # this is last because image can reset values
			self.image = image
		self._data = None

	def fromBytes(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		#print 'Decoding channel at',index
		self.width = io.u32
		self.height = io.u32
		self.name = io.sz754
		self._propertiesDecode_(io)
		self._imageHierarchyPtr = self._pointerDecode_(io)
		self._data = io.data
		return io.index

	def toBytes(self):
		"""
		encode this object to a byte buffer
		"""
		io = IO()
		io.u32 = self.width
		io.u32 = self.height
		io.sz754 = self.name
		io.addBytes(self._propertiesEncode_())
		ih = self._imageHierarchyPtr
		if ih is None:
			ih = 0
		io.addBytes(self._pointerEncode_(ih))
		return io.data

	@property
	def image(self):
		"""
		get a final, compiled image
		"""
		return self.imageHierarchy.image

	@image.setter
	def image(self, image):
		"""
		get a final, compiled image
		"""
		self.width = image.width
		self.height = image.height
		if not self.name and isinstance(image, str):
			# try to use a filename as the name
			self.name = image.rsplit('\\', 1)[-1].rsplit('/', 1)[-1]
		self._imageHierarchy = GimpImageHierarchy(self, image)

	def _forceFullyLoaded(self):
		"""
		make sure everything is fully loaded from the file
		"""
		_ = self.image # make sure the image is loaded so we can delete the hierarchy nonsense
		self._imageHierarchyPtr = None
		self._data = None

	@property
	def imageHierarchy(self):
		"""
		Get teh image hierarchy

		This is mainly used for decoding the image, so
		not much use to you.
		"""
		if self._imageHierarchy is None:
			self._imageHierarchy = GimpImageHierarchy(self)
			self._imageHierarchy.fromBytes(self._data, self._imageHierarchyPtr)
		return self._imageHierarchy

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Name: ' + str(self.name))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append(GimpIOBase.__repr__(self, indent))
		return indent + (('\n' + indent).join(ret))


class GimpImageHierarchy(GimpIOBase):
	"""
	Gets packed pixels from a gimp image

	NOTE: This was originally designed to be a hierarchy, like
		an image pyramid, through in practice they only use the
		top level of the pyramid (64x64) and ignore the rest.
	"""
	def __init__(self, parent, image=None):
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self.bpp = 0 # Number of bytes per pixel given
		self._levelPtrs = []
		self._levels = None
		self._data = None
		if image is not None: # NOTE: can override earlier parameters
			self.image = image

	def fromBytes(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		if not data:
			raise Exception('No data!')
		io = IO(data, index)
		#print 'Decoding channel at',index
		self.width = io.u32
		self.height = io.u32
		self.bpp = io.u32
		if self.bpp < 1 or self.bpp > 4:
			msg = """'Unespected bytes-per-pixel for image data (""" + str(self.bpp) + """).
				Probably means file corruption."""
			raise Exception(msg)
		while True:
			ptr = self._pointerDecode_(io)
			if ptr == 0:
				break
			self._levelPtrs.append(ptr)
		if self._levelPtrs: # remove "dummy" level pointers
			self._levelPtrs = [self._levelPtrs[0]]
		self._data = data
		return io.index

	def toBytes(self):
		"""
		encode this object to a byte buffer
		"""
		dataIO = IO()
		io = IO()
		io.u32 = self.width
		io.u32 = self.height
		io.u32 = self.bpp
		dataIndex = io.index + GimpIOBase._POINTER_SIZE_ * (len(self.levels) + 1)
		for level in self.levels:
			io.addBytes(self._pointerEncode_(dataIndex + dataIO.index)) # TODO: This may be incorrect
			dataIO.addBytes(level.toBytes())
		io.addBytes(self._pointerEncode_(0))
		io.addBytes(dataIO.data)
		return io.data

	@property
	def levels(self):
		"""
		Get the levels within this hierarchy

		Presently hierarchy is not really used by gimp,
		so this returns an array of one item
		"""
		if self._levels is None:
			for ptr in self._levelPtrs:
				l = GimpImageLevel(self)
				l.fromBytes(self._data, ptr)
				self._levels = [l]
		return self._levels

	@property
	def image(self):
		"""
		get a final, compiled image
		"""
		if not self.levels:
			return None
		return self.levels[0].image

	@image.setter
	def image(self, image):
		"""
		set the image
		"""
		self.width = image.width
		self.height = image.height
		if image.mode not in ['L', 'LA', 'RGB', 'RGBA']:
			raise NotImplementedError('Unsupported PIL image type')
		self.bpp = len(image.mode)
		self._levelPtrs = None
		#self._levels = [GimpImageLevel(self, image)]

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('Bytes Per Pixel: ' + str(self.bpp))
		return indent + (('\n' + indent).join(ret))


class GimpImageLevel(GimpIOBase):
	"""
	Gets packed pixels from a gimp image

	This represents a single level in an imageHierarchy
	"""
	def __init__(self, parent):
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self._tiles = None # tile PIL images
		self._image = None

	def fromBytes(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		#print 'Decoding image level at',io.index
		self.width = io.u32
		self.height = io.u32
		if self.width != self.parent.width or self.height != self.parent.height:
			currentSize = '(' + str(self.width) + ',' + str(self.height) + ')'
			expectedSize = '(' + str(self.parent.width) + ',' + str(self.parent.height) + ')'
			msg = ' Usually this implies file corruption.'
			raise Exception('Image data size mismatch. ' + currentSize + '!=' + expectedSize + msg)
		self._tiles = []
		self._image = None
		for y in range(0, self.height, 64):
			for x in range(0, self.width, 64):
				ptr = self._pointerDecode_(io)
				size = (min(self.width - x, 64), min(self.height - y, 64))
				totalBytes = size[0] * size[1] * self.bpp
				if self.doc.compression == 0: # none
					data = io.data[ptr:ptr + totalBytes]
				elif self.doc.compression == 1: # RLE
					data = self._decodeRLE(io.data, size[0] * size[1], self.bpp, ptr)
				elif self.doc.compression == 2: # zip
					data = zlib.decompress(io.data[ptr:ptr + totalBytes +
					24]) # guess how many bytes are needed
				else:
					raise Exception('ERR: unsupported compression mode ' +
					str(self.doc.compression))
				subImage = PIL.Image.frombytes(self.mode, size, bytes(data), decoder_name='raw')
				self._tiles.append(subImage)
		_ = self._pointerDecode_(io) # list ends with nul character
		return io.index

	def toBytes(self):
		"""
		encode this object to a byte buffer
		"""
		dataIO = IO()
		io = IO()
		io.u32 = self.width
		io.u32 = self.height
		#dataIndex = io.index + root._POINTER_SIZE_ * (len(self.tiles) + 1)
		'''
		for tile in self.tiles:
			io.addBytes(self._pointerEncode_(dataIndex + data.index))
			data = tile.tobytes()
			if self.doc.compression == 0: # none
				pass
			elif self.doc.compression == 1: # RLE
				data = self._encodeRLE(data, self.bpp)
			elif self.doc.compression == 2: # zip
				data = zlib.compress(data)
			else:
				raise Exception('ERR: unsupported compression mode ' + str(self.doc.compression))
			dataIO.addBytes(data)
		io.addBytes(self._pointerEncode_(0))
		io.addBytes(dataIO.data)
		return io.data
		'''

	def _decodeRLE(self, data, pixels, bpp, index=0):
		"""
		decode RLE encoded image data
		"""
		ret = [[] for chan in range(bpp)]
		for chan in range(bpp):
			n = 0
			while n < pixels:
				opcode = data[index]
				index += 1
				if opcode >= 0 and opcode <= 126: # a short run of identical bytes
					val = data[index]
					index += 1
					for _ in range(opcode + 1):
						ret[chan].append(val)
						n += 1
				elif opcode == 127: # A long run of identical bytes
					m = data[index]
					index += 1
					b = data[index]
					index += 1
					val = data[index]
					index += 1
					amt = m * 256 + b
					for _ in range(amt):
						ret[chan].append(val)
						n += 1
				elif opcode == 128: # A long run of different bytes
					m = data[index]
					index += 1
					b = data[index]
					index += 1
					amt = m * 256 + b
					for _ in range(amt):
						val = data[index]
						index += 1
						ret[chan].append(val)
						n += 1
				elif opcode >= 129 and opcode <= 255: # a short run of different bytes
					amt = 256 - opcode
					for _ in range(amt):
						val = data[index]
						index += 1
						ret[chan].append(val)
						n += 1
				else:
					print('Unreachable branch', opcode)
					raise Exception()
		# flatten/weave the individual channels into one strream
		flat = bytearray()
		for i in range(pixels):
			for chan in range(bpp):
				flat.append(ret[chan][i])
		return flat

	def _encodeRLE(self, data, bpp):
		"""
		encode image to RLE  image data
		"""
		def countSame(_data, _startIdx):
			"""
			count how many times bytes are identical
			"""
			'''
			l = len(data)
			if idx >= l:
				return 0
			c = data[idx]
			idx = startIdx + 1
			while idx < l and data[idx] == c:
				idx += 1
			return idx - startIdxdef
			'''
			pass

		def countDifferent(data, startIdx):
			"""
			count how many times bytes are different
			"""
			'''
			l = len(data)
			if idx >= l:
				return 0
			c = data[idx]
			idx = startIdx + 1
			while idx < l and data[idx] != c:
				idx += 1
				c = data[idx]
			return idx - startIdx
			'''
			pass

		def rleEncodeChan(data):
			"""
			rle encode a single channel of data
			"""
			ret = []
			idx = 0
			'''
			while idx < len(data):
				nRepeats = countSame(data, 0)
				if nRepeats == 1: # different bytes
					nDifferences = countDifferent(data, 1)
					if nDifferences <= 127: # short run of different bytes
						ret.append(129 + nRepeats - 1)
						ret.append(data[idx])
						idx += nDifferences
					else: # long run of different bytes
						ret.append(128)
						ret.append(floor(nDifferences / 256.0))
						ret.append(nDifferences % 256)
						ret.append(data[idx])
						idx += nDifferences
				elif nRepeats <= 127: # short run of same bytes
					ret.append(nRepeats - 1)
					ret.append(data[idx])
					idx += nRepeats
				else: # long run of same bytes
					ret.append(127)
					ret.append(floor(nRepeats / 256.0))
					ret.append(nRepeats % 256)
					ret.append(data[idx])
					idx += nRepeats
			'''
			return ret

		# if there is only one channel, encode and return it directly
		if bpp == 1:
			return rleEncodeChan(data)
		# split into channels
		dataByChannel = []
		for chan in range(bpp):
			chanData = []
			for index in range(chan, bpp, len(data)):
				chanData.append(data[index])
			dataByChannel.append(chanData)
		# encode each channel
		for index in dataByChannel:
			dataByChannel[index] = rleEncodeChan(dataByChannel[index])
		# join and return
		return ''.join(dataByChannel)

	@property
	def bpp(self):
		""" get bpp """
		return self.parent.bpp

	@property
	def mode(self):
		""" get mode """
		MODES = [None, 'L', 'LA', 'RGB', 'RGBA']
		return MODES[self.bpp]

	@property
	def tiles(self):
		""" get tiles """
		if self._tiles is not None:
			return self._tiles
		if self.image is not None:
			return self._imgToTiles(self.image)
		return None

	def _imgToTiles(self, image):
		"""
		break an image into a series of tiles, each<=64x64
		"""
		ret = []
		for y in range(0, self.height, 64):
			for x in range(0, self.width, 64):
				bounds = (x, y, min(self.width - x, 64), min(self.height - y, 64))
				ret.append(image.crop(bounds))
		return ret

	@property
	def image(self):
		"""
		get a final, compiled image
		"""
		if self._image is None:
			self._image = PIL.Image.new(self.mode, (self.width, self.height), color=None)
			tileNum = 0
			for y in range(0, self.height, 64):
				for x in range(0, self.width, 64):
					subImage = self._tiles[tileNum]
					tileNum += 1
					self._image.paste(subImage, (x, y))
			self._tiles = None # TODO: do I want to keep the tiles for any reason??
		return self._image

	@image.setter
	def image(self, image):
		self._image = image
		self._tiles = None
		self.width = image.width
		self.height = image.height
		self.tiles = None

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		return indent + (('\n' + indent).join(ret))



#!/usr/bin/env python3
"""
A specialized binary file base for Gimp files
"""
import struct
from .binaryIO import IO
from .gimpParasites import GimpParasite


class GimpIOBase:
	"""
	A specialized binary file base for Gimp files
	"""
	COLOR_MODES = ['RGB', 'Grayscale', 'Indexed']
	UNITS = ['Inches', 'Millimeters', 'Points', 'Picas']
	UNITS_TO_MM = [25.4, 1, 127 / 360.0, 127 / 30.0]
	COMPOSITE_MODES = ['Union', 'Clip to backdrop', 'Clip to layer', 'Intersection']
	COMPOSITE_SPACES = ['RGB (linear)', 'RGB (perceptual)', 'LAB']
	TAG_COLORS = ['None', 'Blue', 'Green', 'Yellow', 'Orange', 'Brown', 'Red', 'Violet', 'Gray']
	COMPRESSION_MODES = ['None', 'RLE', 'Zlib', 'Fractal']

	BLEND_MODES = [
	'Normal (legacy)', 'Dissolve (legacy)', 'Behind (legacy)', 'Multiply (legacy)',
	'Screen (legacy)', 'Old broken Overlay', 'Difference (legacy)', 'Addition (legacy)',
	'Subtract (legacy)', 'Darken only (legacy)', 'Lighten only (legacy)', 'Hue (HSV) (legacy)',
	'Saturation (HSV) (legacy)', 'Color (HSL) (legacy)', 'Value (HSV) (legacy)',
	'Divide (legacy)', 'Dodge (legacy)', 'Burn (legacy)', 'Hard Light (legacy)',
	'Soft light (legacy)', 'Grain extract (legacy)', 'Grain merge (legacy)',
	'Color erase (legacy)', 'Overlay', 'Hue (LCH)', 'Chroma (LCH)', 'Color (LCH)',
	'Lightness (LCH)', 'Normal', 'Behind', 'Multiply', 'Screen', 'Difference', 'Addition',
	'Substract', 'Darken only', 'Lighten only', 'Hue (HSV)', 'Saturation (HSV)', 'Color (HSL)',
	'Value (HSV)', 'Divide', 'Dodge', 'Burn', 'Hard light', 'Soft light', 'Grain extract',
	'Grain merge', 'Vivid light', 'Pin light', 'Linear light', 'Hard mix', 'Exclusion',
	'Linear burn', 'Luma/Luminance darken only', 'Luma/Luminance lighten only', 'Luminance',
	'Color erase', 'Erase', 'Merge', 'Split', 'Pass through']

	PROP_END = 0
	PROP_COLORMAP = 1
	PROP_ACTIVE_LAYER = 2
	PROP_ACTIVE_CHANNEL = 3
	PROP_SELECTION = 4
	PROP_FLOATING_SELECTION = 5
	PROP_OPACITY = 6
	PROP_MODE = 7
	PROP_VISIBLE = 8
	PROP_LINKED = 9
	PROP_LOCK_ALPHA = 10
	PROP_APPLY_MASK = 11
	PROP_EDIT_MASK = 12
	PROP_SHOW_MASK = 13
	PROP_SHOW_MASKED = 14
	PROP_OFFSETS = 15
	PROP_COLOR = 16
	PROP_COMPRESSION = 17
	PROP_GUIDES = 18
	PROP_RESOLUTION = 19
	PROP_TATTOO = 20
	PROP_PARASITES = 21
	PROP_UNIT = 22
	PROP_PATHS = 23
	PROP_USER_UNIT = 24
	PROP_VECTORS = 25
	PROP_TEXT_LAYER_FLAGS = 26
	PROP_OLD_SAMPLE_POINTS = 27
	PROP_LOCK_CONTENT = 28
	PROP_GROUP_ITEM = 29
	PROP_ITEM_PATH = 30
	PROP_GROUP_ITEM_FLAGS = 31
	PROP_LOCK_POSITION = 32
	PROP_FLOAT_OPACITY = 33
	PROP_COLOR_TAG = 34
	PROP_COMPOSITE_MODE = 35
	PROP_COMPOSITE_SPACE = 36
	PROP_BLEND_SPACE = 37
	PROP_FLOAT_COLOR = 38
	PROP_SAMPLE_POINTS = 39
	PROP_NUM_PROPS = 40

	def __init__(self, parent):
		self.parent = parent
		self.parasites = []
		self.guidelines = []
		self.itemPath = None
		self.vectors = []
		self.colorMap = []
		self.userUnits = None
		self.samplePoints = []
		self.selected = False
		self.isSelection = False
		self.selectionAttachedTo = None
		self.blendMode = None # one of self.BLEND_MODES
		self.visible = None
		self.isLinked = None
		self.lockAlpha = None
		self.applyMask = None
		self.editingMask = None
		self.showMask = None
		self.showMasked = None
		self.xOffset = None
		self.yOffset = None
		self.compression = None # one of self.COMPRESSION_MODES
		self.horizontalResolution = None
		self.verticalResolution = None
		self.uniqueId = None
		self.units = None # one of self.UNITS
		self.textLayerFlags = None
		self.locked = None
		self.isGroup = None
		self.groupItemFlags = None
		self.positionLocked = None
		self.opacity = None
		self.colorTag = None # one of self.TAG_COLORS
		self.compositeMode = None # one of self.COMPOSITE_MODES
		self.compositeSpace = None # one of self.COMPOSITE_SPACES
		self.blendSpace = None
		self.color = None
		self.vectorsVersion = 0
		self.activeVectorIndex = 0
		self.paths = []

	def getBlendMode(self):
		""" return the blend mode as a string """
		return self.BLEND_MODES[self.blendMode]

	def getCompression(self):
		""" return the compression as a string """
		return self.COMPRESSION_MODES[self.compression]

	def getUnits(self):
		""" return the units as a string """
		return self.UNITS[self.units]

	def getTagColours(self):
		""" return the tag colours as a string """
		return self.TAG_COLORS[self.colorTag]

	def getCompositeModes(self):
		""" return the composite mode as a string """
		return self.COMPOSITE_MODES[abs(self.compositeMode)]

	def getCompositeSpaces(self):
		""" return the composite spaces as a string """
		return self.COMPOSITE_SPACES[abs(self.compositeSpace)]


	@property
	def _POINTER_SIZE_(self):
		"""
		Determine the size of the "pointer" datatype
		based on the document version

		NOTE: prior to version 11, it was 32-bit,
			since then it is 64-bit, thus supporting
			larger image files
		"""
		if self.doc.version >= 11:
			return 64
		return 32

	def _pointerDecode_(self, io):
		if self._POINTER_SIZE_ == 64:
			return io.u64
		return io.u32

	def _pointerEncode_(self, ptr, io=None):
		if not isinstance(ptr, int):
			raise Exception('pointer is wrong type = ' + str(type(ptr)))
		if io is None:
			io = IO()
		if self._POINTER_SIZE_ == 64:
			io.u64 = ptr
		else:
			io.u32 = ptr
		return io.data

	@property
	def doc(self):
		"""
		Get the main document object
		"""
		item = self
		while item.parent != item:
			item = item.parent
		return item

	@property
	def root(self):
		"""
		Get the root of the file object tree
		(Which is the same as self.doc)
		"""
		return self.doc

	@property
	def tattoo(self):
		"""
		This is the gimp nomenclature for the item's unique id
		"""
		return self.uniqueId

	@tattoo.setter
	def tattoo(self, tattoo):
		"""
		This is the gimp nomenclature for the item's unique id
		"""
		self.uniqueId = tattoo

	def _parasitesDecode_(self, data):
		"""
		decode list of parasites
		"""
		index = 0
		while index < len(data):
			p = GimpParasite()
			index = p.fromBytes(data, index)
			self.parasites.append(p)
		return index

	def _parasitesEncode_(self):
		"""
		encode list of parasites
		"""
		io = IO()
		for parasite in self.parasites:
			io.addBytes(parasite.toBytes())
		return io.data

	def _guidelinesDecode_(self, data):
		"""
		decode guidelines
		"""
		index = 0
		while index < len(data):
			position = struct.unpack('>I', data[index:index + 4])[0]
			index += 4
			isVertical = struct.unpack('>B', data[index])[0] == 2
			index += 1
			self.guidelines.append((isVertical, position))

	def _itemPathDecode_(self, data):
		"""
		decode item path
		"""
		index = 0
		path = []
		while index < len(data):
			p = struct.unpack('>I', data[index:index + 4])[0]
			index += 4
			path.append(p)
		self.itemPath = path

	'''
	def _vectorsDecode_(self, data):
		"""
		decode vectors
		"""
		index = 0
		self.vectorsVersion = struct.unpack('>I', data[index:index + 4])[0]
		index += 4
		self.activeVectorIndex = struct.unpack('>I', data[index:index + 4])[0]
		index += 4
		numPaths = struct.unpack('>I', data[index:index + 4])[0]
		index += 4
		for _ in range(numPaths):
			gv = GimpVector(self)
			gv._decode_(data)
			self.vectors.append(gv)
	'''

	@property
	def activeVector(self):
		"""
		get the vector that is currently active
		"""
		return self.vectors[self.activeVectorIndex]

	@property
	def expanded(self):
		"""
		is the group layer expanded
		"""
		return self.groupItemFlags & 0x00000001 > 0

	@expanded.setter
	def expanded(self, expanded):
		"""
		is the group layer expanded
		"""
		if expanded:
			self.groupItemFlags |= 0x00000001
		else:
			self.groupItemFlags &= (~0x00000001)

	def _colormapDecode_(self, data, index=None):
		"""
		:param data: can be bytes or an IO object

		decode colormap/palette
		"""
		ioObj = None
		if isinstance(data, IO):
			ioObj = data
			index = data.index
			data = data.data
		_ = struct.unpack('>I', data[0:4])[0] # number of colors
		index += 4
		colors = []
		while index < len(data):
			r = data[index]
			index += 1
			g = data[index]
			index += 1
			b = data[index]
			index += 1
			colors.append((r, g, b))
		self.colorMap = colors
		if ioObj is not None:
			ioObj.index = index

	def _userUnitsDecode_(self, data):
		"""
		decode a set of user-defined measurement units
		"""
		u = GimpUserUnits()
		#u._decode_(data)
		u.fromBytes(data)
		self.userUnits = u

	def _samplePointsDecode_(self, data):
		"""
		decode a series of points
		"""
		index = 0
		samplePoints = []
		while index < len(data):
			x = struct.unpack('>I', data[index:index + 4])[0]
			index += 4
			y = struct.unpack('>I', data[index:index + 4])[0]
			index += 4
			samplePoints.append((x, y))
		self.samplePoints = samplePoints

	def _propertyDecode_(self, propertyType, data):
		"""
		decode a single property
		"""
		io = IO(data, boolSize=32)
		#print('DECODING PROPERTY',propertyType,len(data))
		if propertyType == self.PROP_COLORMAP:
			self._colormapDecode_(io)
		elif propertyType == self.PROP_ACTIVE_LAYER:
			self.selected = True
		elif propertyType == self.PROP_ACTIVE_CHANNEL:
			self.selected = True
		elif propertyType == self.PROP_SELECTION:
			self.isSelection = True
		elif propertyType == self.PROP_FLOATING_SELECTION:
			self.selectionAttachedTo = io.u32
		elif propertyType == self.PROP_OPACITY:
			self.opacity = io.u32
		elif propertyType == self.PROP_MODE:
			self.blendMode = io.u32
		elif propertyType == self.PROP_VISIBLE:
			self.visible = True
		elif propertyType == self.PROP_LINKED:
			self.isLinked = io.bool
		elif propertyType == self.PROP_LOCK_ALPHA:
			self.lockAlpha = io.bool
		elif propertyType == self.PROP_APPLY_MASK:
			self.applyMask = io.bool
		elif propertyType == self.PROP_EDIT_MASK:
			self.editingMask = io.bool
		elif propertyType == self.PROP_SHOW_MASK:
			self.showMask = io.bool
		elif propertyType == self.PROP_SHOW_MASKED:
			self.showMasked = io.bool
		elif propertyType == self.PROP_OFFSETS:
			self.xOffset = io.i32
			self.yOffset = io.i32
		elif propertyType == self.PROP_COLOR:
			r = io.byte
			g = io.byte
			b = io.byte
			self.color = [r, g, b]
		elif propertyType == self.PROP_COMPRESSION:
			self.compression = io.byte
		elif propertyType == self.PROP_GUIDES:
			self._guidelinesDecode_(data)
		elif propertyType == self.PROP_RESOLUTION:
			self.horizontalResolution = io.float32
			self.verticalResolution = io.float32
		elif propertyType == self.PROP_TATTOO:
			self.uniqueId = data.hex()
		elif propertyType == self.PROP_PARASITES:
			self._parasitesDecode_(data)
		elif propertyType == self.PROP_UNIT:
			self.units = io.u32
		elif propertyType == self.PROP_PATHS:
			_numPaths = io.u32
			'''
			for _ in range(numPaths):
				nRead, path = self._pathDecode_(data[index:])
				self.paths.append(path)
				index += nRead
			'''
		elif propertyType == self.PROP_USER_UNIT:
			self._userUnitsDecode_(data)
		elif propertyType == self.PROP_VECTORS:
			pass
			#self._vectorsDecode_(data)
		elif propertyType == self.PROP_TEXT_LAYER_FLAGS:
			if isinstance(data, bytes):
				self.textLayerFlags = int.from_bytes(data, byteorder='big')
			else:
				self.textLayerFlags = int(data)
		elif propertyType == self.PROP_OLD_SAMPLE_POINTS:
			raise Exception("ERR: old sample points structure not supported")
		elif propertyType == self.PROP_LOCK_CONTENT:
			self.locked = io.bool
		elif propertyType == self.PROP_GROUP_ITEM:
			self.isGroup = True
		elif propertyType == self.PROP_ITEM_PATH:
			self._itemPathDecode_(data)
		elif propertyType == self.PROP_GROUP_ITEM_FLAGS:
			self.groupItemFlags = io.u32
		elif propertyType == self.PROP_LOCK_POSITION:
			self.positionLocked = io.bool
		elif propertyType == self.PROP_FLOAT_OPACITY:
			self.opacity = io.float32
		elif propertyType == self.PROP_COLOR_TAG:
			self.colorTag = io.u32
		elif propertyType == self.PROP_COMPOSITE_MODE:
			self.compositeMode = io.i32
		elif propertyType == self.PROP_COMPOSITE_SPACE:
			self.compositeSpace = io.i32
		elif propertyType == self.PROP_BLEND_SPACE:
			self.blendSpace = io.u32
		elif propertyType == self.PROP_FLOAT_COLOR:
			r = io.float32
			g = io.float32
			b = io.float32
			self.color = [r, g, b]
		elif propertyType == self.PROP_SAMPLE_POINTS:
			self._samplePointsDecode_(data)
		else:
			raise Exception('Unknown property id ' + str(propertyType))
		return io.index

	def _propertyEncode_(self, propertyType):
		"""
		encode a single property

		If the property is the same as the default, or not specified, returns empty array
		"""
		io = IO(boolSize=32)
		if propertyType == self.PROP_COLORMAP:
			if self.colorMap is not None and self.colorMap:
				io.u32 = self.PROP_COLORMAP
				#io.addBytes(self._colormapEncode_())
		elif propertyType == self.PROP_ACTIVE_LAYER:
			if self.selected is not None and self.selected:
				io.u32 = self.PROP_ACTIVE_LAYER
		elif propertyType == self.PROP_ACTIVE_CHANNEL:
			if self.selected is not None and self.selected:
				io.u32 = self.PROP_ACTIVE_LAYER
		elif propertyType == self.PROP_SELECTION:
			if self.isSelection is not None and self.isSelection:
				io.u32 = self.PROP_SELECTION
		elif propertyType == self.PROP_FLOATING_SELECTION:
			if self.selectionAttachedTo is not None:
				io.u32 = self.PROP_FLOATING_SELECTION
				io.u32 = self.selectionAttachedTo
		elif propertyType == self.PROP_OPACITY:
			if self.opacity is not None and not isinstance(self.opacity, float):
				io.u32 = self.PROP_OPACITY
				io.u32 = self.opacity
		elif propertyType == self.PROP_MODE:
			if self.blendMode is not None:
				io.u32 = self.PROP_MODE
				io.u32 = self.blendMode
		elif propertyType == self.PROP_VISIBLE:
			if self.visible is not None and self.visible:
				io.u32 = self.PROP_VISIBLE
		elif propertyType == self.PROP_LINKED:
			if self.isLinked is not None and self.isLinked:
				io.u32 = self.PROP_LINKED
				io.bool = self.isLinked
		elif propertyType == self.PROP_LOCK_ALPHA:
			if self.lockAlpha is not None and self.lockAlpha:
				io.u32 = self.PROP_LOCK_ALPHA
				io.bool = self.lockAlpha
		elif propertyType == self.PROP_APPLY_MASK:
			if self.applyMask is not None:
				io.u32 = self.PROP_APPLY_MASK
				io.bool = self.applyMask
		elif propertyType == self.PROP_EDIT_MASK:
			if self.editingMask is not None and self.editingMask:
				io.u32 = self.PROP_EDIT_MASK
				io.bool = self.editingMask
		elif propertyType == self.PROP_SHOW_MASK:
			if self.showMask is not None and self.showMask:
				io.u32 = self.PROP_SHOW_MASK
				io.bool = self.showMask
		elif propertyType == self.PROP_SHOW_MASKED:
			if self.showMasked is not None:
				io.u32 = self.PROP_SHOW_MASKED
				io.bool = self.showMasked
		elif propertyType == self.PROP_OFFSETS:
			if self.xOffset is not None and self.yOffset is not None:
				io.u32 = self.PROP_OFFSETS
				io.i32 = self.xOffset
				io.i32 = self.yOffset
		elif propertyType == self.PROP_COLOR:
			if self.color is not None and not isinstance(self.color[0],
			float) and not isinstance(self.color[1], float) and not type(
				self.color[2], float):
				io.u32 = self.PROP_COLOR
				io.byte = self.color[0]
				io.byte = self.color[1]
				io.byte = self.color[2]
		elif propertyType == self.PROP_COMPRESSION:
			if self.compression is not None:
				io.u32 = self.PROP_COMPRESSION
				io.u32 = self.compression
		elif propertyType == self.PROP_GUIDES:
			if self.guidelines is not None and self.guidelines:
				io.u32 = self.PROP_GUIDES
				#io.addBytes(self._guidelinesEncode_())
		elif propertyType == self.PROP_RESOLUTION:
			if self.horizontalResolution is not None and self.verticalResolution is not None:
				io.u32 = self.PROP_RESOLUTION
				io.u32 = self.horizontalResolution
				io.i32 = self.verticalResolution
		elif propertyType == self.PROP_TATTOO:
			if self.uniqueId is not None:
				io.u32 = int(self.uniqueId, 16)
		elif propertyType == self.PROP_PARASITES:
			if self.parasites is not None and self.parasites:
				io.u32 = self.PROP_PARASITES
				io.addBytes(self._parasitesEncode_())
		elif propertyType == self.PROP_UNIT:
			if self.units is not None:
				io.u32 = self.PROP_UNIT
				io.u32 = self.units
		elif propertyType == self.PROP_PATHS:
			if self.paths is not None and self.paths:
				io.u32 = self.PROP_PATHS
				io.u32 = len(self.paths)
				'''
				for path in self.paths:
					io.append(self._pathEncode_(path))
				'''
		elif propertyType == self.PROP_USER_UNIT:
			if self.userUnits is not None:
				io.u32 = self.PROP_USER_UNIT
				#io.addBytes(self._userUnitsEncode_())
		elif propertyType == self.PROP_VECTORS:
			if self.vectors is not None and self.vectors:
				io.u32 = self.PROP_VECTORS
				#io.addBytes(self._vectorsEncode_())
		elif propertyType == self.PROP_TEXT_LAYER_FLAGS:
			if self.textLayerFlags is not None:
				io.u32 = self.PROP_TEXT_LAYER_FLAGS
				io.u32 = self.textLayerFlags
		elif propertyType == self.PROP_OLD_SAMPLE_POINTS:
			pass
		elif propertyType == self.PROP_LOCK_CONTENT:
			if self.locked is not None and self.locked:
				io.u32 = self.PROP_LOCK_CONTENT
				io.bool = self.locked
		elif propertyType == self.PROP_GROUP_ITEM:
			if self.isGroup is not None and self.isGroup:
				io.u32 = self.PROP_GROUP_ITEM
		elif propertyType == self.PROP_ITEM_PATH:
			if self.itemPath is not None:
				io.u32 = self.PROP_ITEM_PATH
				#io.addBytes(self._itemPathEncode_())
		elif propertyType == self.PROP_GROUP_ITEM_FLAGS:
			if self.groupItemFlags is not None:
				io.u32 = self.PROP_GROUP_ITEM_FLAGS
				io.u32 = self.groupItemFlags
		elif propertyType == self.PROP_LOCK_POSITION:
			if self.positionLocked is not None and self.positionLocked:
				io.u32 = self.PROP_LOCK_POSITION
				io.bool = self.positionLocked
		elif propertyType == self.PROP_FLOAT_OPACITY:
			if self.opacity is not None and isinstance(self.opacity, float):
				io.u32 = self.PROP_FLOAT_OPACITY
				io.float32 = self.opacity
		elif propertyType == self.PROP_COLOR_TAG:
			if self.colorTag is not None:
				io.u32 = self.PROP_COLOR_TAG
				io.u32 = self.colorTag
		elif propertyType == self.PROP_COMPOSITE_MODE:
			if self.compositeMode is not None:
				io.u32 = self.PROP_COMPOSITE_MODE
				io.i32 = self.compositeMode
		elif propertyType == self.PROP_COMPOSITE_SPACE:
			if self.compositeSpace is not None:
				io.u32 = self.PROP_COMPOSITE_SPACE
				io.i32 = self.compositeSpace
		elif propertyType == self.PROP_BLEND_SPACE:
			if self.blendSpace is not None:
				io.u32 = self.PROP_BLEND_SPACE
				io.u32 = self.blendSpace
		elif propertyType == self.PROP_FLOAT_COLOR:
			if self.color is not None and isinstance(self.color[0], float) or isinstance(
				self.color[1], float) or isinstance(self.color[2], float):
				io.u32 = self.PROP_FLOAT_COLOR
				io.float32 = self.color[0]
				io.float32 = self.color[1]
				io.float32 = self.color[2]
		elif propertyType == self.PROP_SAMPLE_POINTS:
			if self.samplePoints is not None and self.samplePoints:
				io.u32 = self.PROP_SAMPLE_POINTS
				#self.addBytes(self._samplePointsEncode_())
		else:
			raise Exception('Unknown property id ' + str(propertyType))
		return io.data

	def _propertiesDecode_(self, io):
		"""
		decode a list of properties
		"""
		while True:
			try:
				propertyType = io.u32
				dataLength = io.u32
			except struct.error: # end of data, so that's that.
				break
			if propertyType == 0:
				break
			self._propertyDecode_(propertyType, io.getBytes(dataLength))
		return io.index

	def _propertiesEncode_(self):
		"""
		encode a list of properties
		"""
		io = IO()
		for propertyType in range(1, self.PROP_NUM_PROPS):
			moData = self._propertyEncode_(propertyType)
			if moData:
				io.addBytes(moData)
		return io.data

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		if self.itemPath is not None:
			ret.append('Item Path: ' + str(self.itemPath))
		if self.selected is not None:
			ret.append('Selected: ' + str(self.selected))
		if self.isSelection is not None:
			ret.append('is Selection: ' + str(self.isSelection))
		if self.selectionAttachedTo is not None:
			ret.append('Selection Attached To: ' + str(self.selectionAttachedTo))
		if self.blendMode is not None:
			ret.append('Blend Mode: ' + self.getBlendMode())
		if self.visible is not None:
			ret.append('Visible: ' + str(self.visible))
		if self.isLinked is not None:
			ret.append('Linked: ' + str(self.isLinked))
		if self.lockAlpha is not None:
			ret.append('Alpha Locked: ' + str(self.lockAlpha))
		if self.applyMask is not None:
			ret.append('Apply Mask: ' + str(self.applyMask))
		if self.editingMask is not None:
			ret.append('Editing Mask: ' + str(self.editingMask))
		if self.showMask is not None:
			ret.append('Show Mask: ' + str(self.showMask))
		if self.showMasked is not None:
			ret.append('Show Masked: ' + str(self.showMasked))
		if self.xOffset is not None:
			ret.append('Offset: ' + str(self.xOffset) + ' x ' + str(self.yOffset))
		if self.compression is not None:
			ret.append('Compression: ' + self.getCompression())
		if self.horizontalResolution is not None:
			res = str(self.horizontalResolution) + 'ppi x ' + str(self.verticalResolution) + 'ppi'
			ret.append('Resolution: ' + res)
		if self.uniqueId is not None:
			ret.append('Unique ID (tattoo): ' + str(self.uniqueId))
		if self.units is not None:
			ret.append('Units: ' + self.getUnits())
		if self.textLayerFlags is not None:
			ret.append('Text Layer Flags: ' + str(self.textLayerFlags))
		if self.locked is not None:
			ret.append('Locked: ' + str(self.locked))
		if self.isGroup is not None:
			ret.append('Is Group: ' + str(self.isGroup))
		if self.groupItemFlags is not None:
			ret.append('Group Items Flag: ' + str(self.groupItemFlags))
		if self.positionLocked is not None:
			ret.append('Position Locked: ' + str(self.positionLocked))
		if self.opacity is not None:
			ret.append('Opacity: ' + str(self.opacity))
		if self.colorTag is not None:
			ret.append('Tag Color: ' + self.getTagColours())
		if self.compositeMode is not None:
			auto = ('Auto ' if self.compositeMode < 0 else '') # negative values are "Auto"
			ret.append('Composite Mode: ' + auto + self.getCompositeModes())
		if self.compositeSpace is not None:
			auto = ('Auto ' if self.compositeSpace < 0 else '') # negative values are "Auto"
			ret.append('Composite Space: ' + auto +
			self.getCompositeSpaces())
		if self.blendSpace is not None:
			ret.append('Blend Space: ' + str(self.blendSpace))
		if self.color is not None:
			ret.append('Color: (' + str(self.color[0]) + ',' + str(self.color[1]) + ',' +
			str(self.color[2]) + ')')
		if self.userUnits is not None:
			ret.append('User Units: ')
			ret.append(self.userUnits.__repr__(indent + '\t'))
		if self.parasites:
			ret.append('Parasites: ')
			for item in self.parasites:
				ret.append(item.__repr__(indent + '\t'))
		if self.guidelines:
			ret.append('Guidelines: ')
			for item in self.guidelines:
				ret.append(item.__repr__(indent + '\t'))
		if self.samplePoints:
			ret.append('Sample Points: ')
			for item in self.samplePoints:
				ret.append('(' + str(item[0]) + ',' + str(item[1]) + ')')
		if self.vectors:
			ret.append('Vectors: ')
			for item in self.vectors:
				ret.append(item.__repr__(indent + '\t'))
		if self.colorMap:
			ret.append('Color Map: ')
			for i in range(len((self.colorMap))):
				color = self.colorMap[i]
				ret.append(i + ': (' + str(color[0]) + ',' + str(color[1]) + ',' + str(color[2]) +
				')')
		return indent + (('\n' + indent).join(ret))


class GimpUserUnits:
	"""
	user-defined measurement units
	"""
	def __init__(self):
		self.factor = 0
		self.numDigits = 0
		self.id = ''
		self.symbol = ''
		self.abbrev = ''
		self.sname = ''
		self.pname = ''

	def fromBytes(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		self.factor = io.float32
		self.numDigits = io.u32
		self.id = io.sz754
		self.symbol = io.sz754
		self.abbrev = io.sz754
		self.sname = io.sz754
		self.pname = io.sz754
		return io.index

	def toBytes(self):
		"""
		convert this object to raw bytes
		"""
		io = IO()
		io.float32 = self.factor
		io.u32 = self.numDigits
		io.sz754 = self.id
		io.sz754 = self.symbol
		io.sz754 = self.abbrev
		io.sz754 = self.sname
		io.sz754 = self.pname
		return io.data

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Factor: ' + str(self.factor))
		ret.append('Num Digits: ' + str(self.numDigits))
		ret.append('ID: ' + str(self.id))
		ret.append('Symbol: ' + str(self.symbol))
		ret.append('Abbreviation: ' + str(self.abbrev))
		ret.append('Singular Name: ' + str(self.sname))
		ret.append('Plural Name: ' + str(self.pname))
		return indent + (('\n' + indent).join(ret))



#!/usr/bin/env python3
"""
Pure python implementation of the gimp gbr brush format
"""
import argparse
import PIL.Image
from .binaryIO import IO

class GimpGbrBrush:
	"""
	Pure python implementation of the gimp gbr brush format

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/gbr.txt
	"""

	COLOR_MODES = [None, 'L', 'LA', 'RGB', 'RGBA'] # only L or RGB allowed

	def __init__(self, filename=None):
		self.filename = None
		self.version = 2
		self.width = 0
		self.height = 0
		self.bpp = 1
		self.mode = self.COLOR_MODES[self.bpp]
		self.name = ''
		self.rawImage = None
		self.spacing = 0
		if filename is not None:
			self.load(filename)

	def load(self, filename):
		"""
		load a gimp file

		:param filename: can be a file name or a file-like object
		"""
		if hasattr(filename, 'read'):
			self.filename = filename.name
			f = filename
		else:
			self.filename = filename
			f = open(filename, 'rb')
		data = f.read()
		f.close()
		self._decode_(data)

	def _decode_(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		headerSize = io.u32
		self.version = io.u32
		if self.version != 2:
			raise Exception('ERR: unknown brush version ' + str(self.version))
		self.width = io.u32
		self.height = io.u32
		self.bpp = io.u32 # only allows grayscale or RGB
		self.mode = self.COLOR_MODES[self.bpp]
		magic = io.getBytes(4)
		if magic.decode('ascii') != 'GIMP':
			raise Exception('"' + magic.decode('ascii') + '" ' + str(index) +
		" File format error.  Magic value mismatch.")
		self.spacing = io.u32
		nameLen = headerSize - io.index
		self.name = io.getBytes(nameLen).decode('UTF-8')
		self.rawImage = io.getBytes(self.width * self.height * self.bpp)
		return io.index

	def toBytes(self):
		"""
		encode this object to byte array
		"""
		io = IO()
		io.u32 = 28 + len(self.name)
		io.u32 = self.version
		io.u32 = self.width
		io.u32 = self.height
		io.u32 = self.bpp
		io.addBytes('GIMP')
		io.u32 = self.spacing
		io.addBytes(self.name)
		io.addBytes(self.rawImage)
		return io.data

	@property
	def size(self):
		""" Get the size """
		return (self.width, self.height)

	@property
	def image(self):
		"""	get a final, compiled image """
		if self.rawImage is None:
			return None
		return PIL.Image.frombytes(self.mode, self.size, self.rawImage, decoder_name='raw')

	def save(self, toFilename=None, toExtension=None):
		"""	save this gimp image to a file """
		asImage = False
		if toExtension is None:
			if toFilename is not None:
				toExtension = toFilename.rsplit('.', 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if toExtension is not None and toExtension != 'gbr':
			asImage = True
		if asImage:
			self.image.save(toFilename)
			self.image.close()
		else:
			if not hasattr(toFilename, 'write'):
				f = open(toFilename, 'wb')
			f.write(self.toBytes())
			f.close()

	def __repr__(self, indent=''):
		"""	Get a textual representation of this object """
		ret = []
		if self.filename is not None:
			ret.append('Filename: ' + self.filename)
		ret.append('Name: ' + str(self.name))
		ret.append('Version: ' + str(self.version))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('Spacing: ' + str(self.spacing))
		ret.append('BPP: ' + str(self.bpp))
		ret.append('Mode: ' + str(self.mode))
		return ('\n' + indent).join(ret)


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("gimpGbrBrush.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--dump", action="store_true",
	help="dump info about this file")
	group.add_argument("--show", action="store_true",
	help="show the brush image")
	group.add_argument("--save", action="store",
	help="save out the brush image")
	args = parser.parse_args()

	gimpDocument = GimpGbrBrush(args.xcfdocument)

	if args.dump:
		print(gimpDocument)
	if args.show:
		gimpDocument.image.show()
	if args.save:
		gimpDocument.image.save(args.save)



#!/usr/bin/env python3
"""
Pure python implementation of the OLD gimp gpb brush format
"""
import argparse
from .binaryIO import IO
from .gimpGbrBrush import GimpGbrBrush
from .gimpPatPattern import GimpPatPattern


class GimpGpbBrush:
	"""
	Pure python implementation of the OLD gimp gpb brush format

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/vbr.txt
	"""
	def __init__(self, filename):
		self.brush = GimpGbrBrush()
		self.pattern = GimpPatPattern()
		if hasattr(filename, 'read'):
			self.filename = filename.name
		else:
			self.filename = filename

	def load(self, filename):
		"""
		load a gimp file

		:param filename: can be a file name or a file-like object
		"""
		if hasattr(filename, 'read'):
			self.filename = filename.name
			f = filename
		else:
			self.filename = filename
			f = open(filename, 'rb')
		data = f.read()
		f.close()
		self._decode_(data)

	def _decode_(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		index = self.brush._decode_(data, index)
		#index = self.pattern._decode_(data, index)
		return index

	def toBytes(self):
		""" encode this object to a byte array """
		io = IO()
		io.addBytes(self.brush.toBytes())
		io.addBytes(self.pattern.toBytes())
		return io.data

	def save(self, toFilename=None):
		""" save this gimp image to a file """
		if not hasattr(toFilename, 'write'):
			f = open(toFilename, 'wb')
		f.write(self.toBytes())

	def __repr__(self, indent=''):
		""" Get a textual representation of this object """
		ret = []
		if self.filename is not None:
			ret.append('Filename: ' + self.filename)
		ret.append(self.brush.__repr__(indent + '\t'))
		ret.append(self.pattern.__repr__(indent + '\t'))
		return ('\n' + indent).join(ret)


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("gimpGbrBrush.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	parser.add_argument("--dump", action="store_true",
	help="dump info about this file")
	args = parser.parse_args()

	gimpDocument = GimpGbrBrush(args.xcfdocument)

	if args.dump:
		print(gimpDocument)



#!/usr/bin/env python3
"""
Pure python implementation of the gimp xcf file format

Currently supports:
	Loading xcf files
	Getting image hierarchy and info
	Getting image for each layer (PIL image)
Currently not supporting:
	Saving
	Programatically alter documents (add layer, etc)
	Rendering a final, compositied image
"""
import argparse
import copy

import PIL.ImageGrab
from PIL import Image
from blendmodes.blend import blendLayers, BlendType
from .binaryIO import IO
from .gimpIOBase import GimpIOBase
from .gimpImageInternals import GimpChannel, GimpImageHierarchy


class GimpLayer(GimpIOBase):
	"""
	Represents a single layer in a gimp image
	"""

	COLOR_MODES = [
	'RGB color without alpha', 'RGB color with alpha', 'Grayscale without alpha',
	'Grayscale with alpha', 'Indexed without alpha', 'Indexed with alpha']
	PIL_MODE_TO_LAYER_MODE = {'L': 2, 'LA': 3, 'RGB': 0, 'RGBA': 1}

	def __init__(self, parent, name=None, image=None):
		GimpIOBase.__init__(self, parent)
		self.width = 0
		self.height = 0
		self.colorMode = 0
		self.name = name
		self._imageHierarchy = None
		self._imageHierarchyPtr = None
		self._mask = None
		self._maskPtr = None
		self._data = None
		if image is not None:
			self.image = image # done last as it resets some of the above defaults

	def _decode_(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		#print 'Decoding Layer at',index
		self.width = io.u32
		self.height = io.u32
		self.colorMode = io.u32 # one of self.COLOR_MODES
		self.name = io.sz754
		self._propertiesDecode_(io)
		self._imageHierarchyPtr = self._pointerDecode_(io)
		self._maskPtr = self._pointerDecode_(io)
		self._mask = None
		self._data = data
		return io.index

	def toBytes(self):
		"""
		encode to byte array
		"""
		dataAreaIO = IO()
		io = IO()
		io.u32 = self.width
		io.u32 = self.height
		io.u32 = self.colorMode
		io.sz754 = self.name
		dataAreaIndex = io.index + self._POINTER_SIZE_ * 2
		io.addBytes(self._pointerEncode_(dataAreaIndex))
		dataAreaIO.addBytes(self._propertiesEncode_())
		io.addBytes(self._pointerEncode_(dataAreaIndex))
		if self.mask is not None:
			dataAreaIO.addBytes(self.mask.toBytes())
		io.addBytes(self._pointerEncode_(dataAreaIndex + dataAreaIO.index))
		io.addBytes(dataAreaIO)
		return io.data

	@property
	def mask(self):
		"""
		Get the layer mask
		"""
		if self._mask is None and self._maskPtr is not None and self._maskPtr != 0:
			self._mask = GimpChannel(self)
			self._mask.fromBytes(self._data, self._maskPtr)
		return self._mask

	@property
	def image(self):
		"""
		get the layer image

		NOTE: can return None!
		"""
		if self.imageHierarchy is None:
			return None
		return self.imageHierarchy.image

	@image.setter
	def image(self, image):
		"""
		set the layer image

		NOTE: resets layer width, height, and colorMode
		"""
		self.height = image.height
		self.width = image.width
		if image.mode not in self.PIL_MODE_TO_LAYER_MODE:
			raise NotImplementedError('No way of handlng PIL image mode "' + image.mode + '"')
		self.colorMode = self.PIL_MODE_TO_LAYER_MODE[image.mode]
		if not self.name and isinstance(image, str):
			# try to use a filename as the name
			self.name = image.rsplit('\\', 1)[-1].rsplit('/', 1)[-1]
		self.imageHierarchy = GimpImageHierarchy(self)
		self.imageHierarchy.image = image

	@property
	def imageHierarchy(self):
		"""
		Get the image hierarchy objects

		This is mainly needed for deciphering image, and therefore,
		of little use to you, the user.

		NOTE: can return None if it has been fully read into an image
		"""
		if self._imageHierarchy is None and self._imageHierarchyPtr > 0:
			self._imageHierarchy = GimpImageHierarchy(self)
			self._imageHierarchy.fromBytes(self._data, self._imageHierarchyPtr)
		return self._imageHierarchy

	def _forceFullyLoaded(self):
		"""
		make sure everything is fully loaded from the file
		"""
		if self.mask is not None:
			self.mask._forceFullyLoaded()
		_ = self.image # make sure the image is loaded so we can delete the hierarchy nonsense
		self._imageHierarchy = None
		self._data = None

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Name: ' + str(self.name))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('colorMode: ' + self.COLOR_MODES[self.colorMode])
		ret.append(GimpIOBase.__repr__(self, indent))
		m = self.mask
		if m is not None:
			ret.append('Mask:')
			ret.append(m.__repr__(indent + '\t'))
		return indent + (('\n' + indent).join(ret))


class Precision:
	"""
	Since the precision code is so unusual, I decided to create a class
	to parse it.
	"""
	def __init__(self):
		self.bits = 8
		self.gamma = True
		self.numberFormat = int

	def decode(self, gimpVersion, io):
		"""
		decode the precision code from the file
		"""
		if gimpVersion < 4:
			self.bits = 8
			self.gamma = True
			self.numberFormat = int
		else:
			code = io.u32
			if gimpVersion == 4:
				self.gamma = (True, True, False, False, False)[code]
				self.bits = (8, 16, 32, 16, 32)[code]
				self.numberFormat = (int, int, int, float, float)[code]
			elif gimpVersion in (5, 6):
				self.gamma = (code % 100 != 0)
				code = int(code / 100)
				self.bits = (8, 16, 32, 16, 32)[code]
				self.numberFormat = (int, int, int, float, float)[code]
			else: # gimpVersion 7 or above
				self.gamma = (code % 100 != 0)
				code = int(code / 100)
				self.bits = (8, 16, 32, 16, 32, 64)[code]
				self.numberFormat = (int, int, int, float, float, float)[code]

	def encode(self, gimpVersion, io):
		"""
		encode this to the file

		NOTE: will not mess with development versions 5 or 6
		"""
		if gimpVersion < 4:
			if self.bits != 8 or not (self.gamma) or self.numberFormat != int:
				raise Exception('Illegal precision (' + str(self) + ') for gimp version ' +
				str(gimpVersion))
		else:
			if gimpVersion == 4:
				if self.bits == 64:
					raise Exception('Illegal precision (' + str(self) + ') for gimp version ' +
					str(gimpVersion))
				if self.numberFormat == int:
					code = (8, 16, 32).index(self.bits)
				else:
					code = (16, 32).index(self.bits) + 2
				code = code * 100
				if self.gamma:
					code += 50
			elif gimpVersion in (5, 6):
				raise NotImplementedError('Cannot save to gimp developer version ' +
				str(gimpVersion))
			else: # version 7 or above
				if self.numberFormat == int:
					code = (8, 16, 32).index(self.bits)
				else:
					code = (16, 32, 64).index(self.bits) + 2
				code = code * 100
				if self.gamma:
					code += 50
			io.u32 = code

	def requiredGimpVersion(self):
		"""
		return the lowest gimp version that supports this precision
		"""
		if self.bits == 8 and self.gamma and self.numberFormat == int:
			return 0
		if self.bits == 64:
			return 7
		return 4

	def __repr__(self):
		ret = []
		ret.append(str(self.bits) + "-bit")
		ret.append('gamma' if self.gamma else 'linear')
		ret.append('integer' if self.numberFormat is int else float)
		return ' '.join(ret)


class GimpDocument(GimpIOBase):
	"""
	Pure python implementation of the gimp file format

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/xcf.txt
	"""
	def __init__(self, fileName=None):
		GimpIOBase.__init__(self, self)
		self.dirty = False # a file-changed indicator.  # TODO: Not fully implemented.
		self._layers = None
		self._layerPtr = []
		self.channels = []
		self._channelPtr = []
		self.version = None
		self.width = 0
		self.height = 0
		self.baseColorMode = 0
		self.precision = None # Precision object
		self._data = None
		if fileName is not None:
			self.load(fileName)

	def load(self, fileName):
		"""
		load a gimp file

		:param filename: can be a file name or a file-like object
		"""
		if hasattr(fileName, 'read'):
			self.filename = fileName.name
			f = fileName
		else:
			self.filename = fileName
			f = open(fileName, 'rb')
		data = f.read()
		f.close()
		self._decode_(data)

	def _decode_(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		if io.getBytes(9) != "gimp xcf ".encode('ascii'):
			raise Exception('Not a valid GIMP file')
		version = io.cString
		if version == 'file':
			self.version = 0
		else:
			self.version = int(version[1:])
		self.width = io.u32
		self.height = io.u32
		self.baseColorMode = io.u32
		self.precision = Precision()
		self.precision.decode(self.version, io)
		self._propertiesDecode_(io)
		self._layerPtr = []
		self._layers = []
		while True:
			ptr = self._pointerDecode_(io)
			if ptr == 0:
				break
			self._layerPtr.append(ptr)
			l = GimpLayer(self)
			l._decode_(io.data, ptr)
			self._layers.append(l)
		self._channelPtr = []
		self.channels = []
		while True:
			ptr = self._pointerDecode_(io)
			if ptr == 0:
				break
			self._channelPtr.append(ptr)
			c = GimpChannel(self)
			c.fromBytes(io.data, ptr)
			self.channels.append(c)
		return io.index

	def toBytes(self):
		"""
		encode to a byte array
		"""
		io = IO()
		io.addBytes("gimp xcf ")
		io.addBytes(str(self.version) + '\0')
		io.u32 = self.width
		io.u32 = self.height
		io.u32 = self.baseColorMode
		if self.precision is None:
			self.precision = Precision()
		self.precision.encode(self.version, io)
		io.addBytes(self._propertiesEncode_())
		dataAreaIdx = io.index + self._POINTER_SIZE_ * (len(self.layers) + len(self.channels))
		dataAreaIo = IO()
		for l in self.layers:
			io.pointer = dataAreaIdx + dataAreaIo.index
			dataAreaIo.addBytes(l.toBytes())
		for channel in self.channels:
			io.pointer = dataAreaIdx + dataAreaIo.index
			dataAreaIo.addBytes(channel.toBytes())
		return io.data

	def _forceFullyLoaded(self):
		"""
		make sure everything is fully loaded from the file
		"""
		for l in self.layers:
			l._forceFullyLoaded()
		for chan in self.channels:
			chan._forceFullyLoaded()
		# no longer try to get the data from file
		self._layerPtr = None
		self._channelPtr = None
		self._data = None

	@property
	def layers(self):
		"""
		Decode the image's layers if necessary

		TODO: need to do the same thing with self.Channels
		"""
		if self._layers is None:
			self._layers = []
			for _ptr in self._layerPtr:
				l = GimpLayer(self)
				#l.fromBytes(self._data, ptr)
				self._layers.append(l)
			# add a reference back to this object so it doesn't go away while array is in use
			self._layers.parent = self
			# override some internal methods so we can do more with them
			self._layers._actualDelitem_ = self._layers.__delitem__
			self._layers.__delitem__ = self.deleteLayer
			self._layers._actualSetitem_ = self._layers.__delitem__
			self._layers.__setitem__ = self.setLayer
		return self._layers

	def getLayer(self, index):
		"""
		return a given layer
		"""
		return self.layers[index]

	def setLayer(self, _index, _l):
		"""
		assign to a given layer
		"""
		self._forceFullyLoaded()
		self.dirty = True
		self._layerPtr = None # no longer try to use the pointers to get data
		#self.layers._actualSetitem(index, l)

	def newLayer(self, name, image, index=-1):
		"""
		create a new layer based on a PIL image

		:param name: a name for the new layer
		:param index: where to insert the new layer (default=top)
		:return: newly created GimpLayer object
		"""
		l = GimpLayer(self, name, image)
		self.insertLayer(l, index)
		return layer

	def newLayerFromClipboard(self, name='pasted', index=-1):
		"""
		Create a new image from the system clipboard.

		:param name: a name for the new layer (default="pasted")
		:param index: where to insert the new layer (default=top)
		:return: newly created GimpLayer object

		NOTE: requires pillow PIL implementation
		NOTE: only works on OSX and Windows
		"""
		image = PIL.ImageGrab.grabclipboard()
		return self.newLayer(name, image, index)

	def addLayer(self, l):
		"""
		append a layer object to the document

		:param layer: the new layer to append
		"""
		self.insertLayer(l, -1)

	def appendLayer(self, l):
		"""
		append a layer object to the document

		:param layer: the new layer to append
		"""
		self.insertLayer(l, -1)

	def insertLayer(self, l, index=-1):
		"""
		insert a layer object at a specific position

		:param layer: the new layer to insert
		:param index: where to insert the new layer (default=top)
		"""
		self.layers.insert(index, l)

	def deleteLayer(self, index):
		"""
		delete a layer
		"""
		self.__delitem__(index)

	# make this class act like this class is an array of layers
	def __len__(self):
		return len(self.layers)

	def __getitem__(self, index):
		return self.layers[index]

	def __setitem__(self, index, l):
		self.setLayer(index, l)

	def __delitem__(self, index):
		self.deleteLayer(index)

	def __inc__(self, amt):
		self.appendLayer(amt)
		return self

	@property
	def image(self):
		"""
		get a final, compiled image
		"""
		# We need to preprocess the layers to sort them into a list of layers
		# and groups. Where a group is a list of layers

		# Example Layers [layer, layer, group, layer]
		# Example Group [Layer(Group), [layer, layer, ...]]
		layers = self.layers[:] # Copy the attribute rather than write to it
		layersOut = []
		index = 0
		while index < len(layers):
			layerOrGroup = layers[index]
			if layerOrGroup.isGroup:
				elem = [layerOrGroup, []]
				index += 1
				while layers[index].itemPath is not None:
					layerCopy = copy.deepcopy(layers[index])
					layerCopy.xOffset -= layerOrGroup.xOffset
					layerCopy.yOffset -= layerOrGroup.yOffset
					elem[1].append(layerCopy)
					layers.pop(index)
				layersOut.append(elem)
			else:
				layersOut.append(layerOrGroup)
				index += 1
		return flattenAll(layersOut, (self.width, self.height))

	def save(self, toFilename=None):
		"""
		save this gimp image to a file
		"""
		self._forceFullyLoaded()
		if toFilename is None:
			toFilename = self.filename
		if not hasattr(toFilename, 'write'):
			f = open(toFilename, 'wb')
		f.write(self.toBytes())
		self.dirty = False

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		if self.filename is not None:
			ret.append('Filename: ' + self.filename)
		ret.append('Version: ' + str(self.version))
		ret.append('Size: ' + str(self.width) + ' x ' + str(self.height))
		ret.append('Base Color Mode: ' + self.COLOR_MODES[self.baseColorMode])
		ret.append('Precision: ' + str(self.precision))
		ret.append(GimpIOBase.__repr__(self))
		if self._layerPtr:
			ret.append('Layers: ')
			for l in self.layers:
				ret.append(l.__repr__('\t'))
		if self._channelPtr:
			ret.append('Channels: ')
			for ch in self.channels:
				ret.append(ch.__repr__('\t'))
		return '\n'.join(ret)

def blendModeLookup(blendmode, blendLookup, default=BlendType.NORMAL):
	""" Get the blendmode from a lookup table """
	if blendmode not in blendLookup:
		print("WARNING " + str(blendmode) + " is not currently supported!")
		return default
	return blendLookup[blendmode]

def rasterImageOffset(image, size, offsets=(0, 0)):
	""" Rasterise an image with offset to a given size"""
	imageOffset = Image.new("RGBA", size)
	imageOffset.paste(image.convert("RGBA"), offsets, image.convert("RGBA"))
	return imageOffset

def flattenLayerOrGroup(layerOrGroup, imageDimensions, flattenedSoFar=None,
ignoreHidden=True):
	"""Flatten a layer or group on to an image of what has already been
	flattened
	Args:
		layerOrGroup (Layer|Group): A layer or a group of layers
		imageDimensions ((int, int)): size of the image
		flattenedSoFar (PIL.Image, optional): the image of what has already
		been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.
	Returns:
		PIL.Image: Flattened image
	"""
	blendLookup = {0: BlendType.NORMAL, 3: BlendType.MULTIPLY,
	4: BlendType.SCREEN, 5: BlendType.OVERLAY, 6: BlendType.DIFFERENCE,
	7: BlendType.ADDITIVE, 8: BlendType.NEGATION, 9: BlendType.DARKEN,
	10: BlendType.LIGHTEN, 11: BlendType.HUE, 12: BlendType.SATURATION,
	13: BlendType.COLOUR, 14: BlendType.LUMINOSITY, 15: BlendType.DIVIDE,
	16: BlendType.COLOURDODGE, 17: BlendType.COLOURBURN,
	18: BlendType.HARDLIGHT, 19: BlendType.SOFTLIGHT, 20: BlendType.GRAINEXTRACT,
	21: BlendType.GRAINMERGE, 23: BlendType.OVERLAY, 24: BlendType.HUE,
	25: BlendType.SATURATION, 26: BlendType.COLOUR, 27: BlendType.LUMINOSITY,
	28: BlendType.NORMAL, 30: BlendType.MULTIPLY, 31: BlendType.SCREEN,
	32: BlendType.DIFFERENCE, 33: BlendType.ADDITIVE, 34: BlendType.NEGATION,
	35: BlendType.DARKEN, 36: BlendType.LIGHTEN, 37: BlendType.HUE,
	38: BlendType.SATURATION, 39: BlendType.COLOUR, 40: BlendType.LUMINOSITY,
	41: BlendType.DIVIDE, 42: BlendType.COLOURDODGE,
	43: BlendType.COLOURBURN, 44: BlendType.HARDLIGHT, 45: BlendType.SOFTLIGHT,
	46: BlendType.GRAINEXTRACT, 47: BlendType.GRAINMERGE,
	48: BlendType.VIVIDLIGHT, 49: BlendType.PINLIGHT, 52: BlendType.EXCLUSION}

	if isinstance(layerOrGroup, list):
		if ignoreHidden and not layerOrGroup[0].visible:
			foregroundRaster = Image.new("RGBA", imageDimensions)
		else: # A group is a list of layers
			# (see flattenAll)
			foregroundRaster = rasterImageOffset(flattenAll(layerOrGroup[1],
			imageDimensions, ignoreHidden), imageDimensions, (layerOrGroup[0].xOffset,
			layerOrGroup[0].yOffset))
		if flattenedSoFar is None:
			return foregroundRaster
		return blendLayers(flattenedSoFar, foregroundRaster,
		blendModeLookup(layerOrGroup[0].blendMode, blendLookup),
		layerOrGroup[0].opacity)

	if ignoreHidden and not layerOrGroup.visible:
		foregroundRaster = Image.new("RGBA", imageDimensions)
	else:
		# Get a raster image and apply blending
		foregroundRaster = rasterImageOffset(layerOrGroup.image, imageDimensions,
		(layerOrGroup.xOffset, layerOrGroup.yOffset))
	if flattenedSoFar is None:
		return foregroundRaster
	return blendLayers(flattenedSoFar, foregroundRaster,
	blendModeLookup(layerOrGroup.blendMode, blendLookup),
	layerOrGroup.opacity)


def flattenAll(layers, imageDimensions, ignoreHidden=True):
	"""Flatten a list of layers and groups

	Note the bottom layer is at the end of the list

	Args:
		layers ([Layer|Group]): A list of layers and groups
		imageDimensions ((int, int)): size of the image
		been flattened. Defaults to None.
		ignoreHidden (bool, optional): ignore layers that are hidden. Defaults
		to True.
	Returns:
		PIL.Image: Flattened image
	"""
	end = len(layers) - 1
	flattenedSoFar = flattenLayerOrGroup(layers[end], imageDimensions, ignoreHidden=ignoreHidden)
	for l in range(end - 1, -1, -1):
		flattenedSoFar = flattenLayerOrGroup(layers[l], imageDimensions,
		flattenedSoFar=flattenedSoFar, ignoreHidden=ignoreHidden)
	return flattenedSoFar


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("gimpGbrBrush.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--dump", action="store_true",
	help="dump info about this file")
	group.add_argument("--showLayer", action="store_true",
	help="show layer(s) (use * for all)")
	group.add_argument("--saveLayer", action="store_true",
	help="save layer(s) out to file")
	group.add_argument("--save", action="store",
	help="save out the brush image")
	args = parser.parse_args()

	gimpDocument = GimpDocument(args.xcfdocument)

	if args.dump:
		print(gimpDocument)
	if args.showLayer:
		if args.showLayer == '*':
			for layer in range(len(gimpDocument.layers)):
				im = gimpDocument.layers[layer].image
				showLayer(im, layer)
		else:
			im = gimpDocument.layers[int(args.showLayer)].image
			showLayer(im, int(args.showLayer))
	if args.saveLayer:
		layer = args.saveLayer.split(',', 1)
		if len(layer) > 1:
			filename = layer[1]
		else:
			filename = 'layer *.png'
		layer = args.saveLayer[0]
		if layer == '*':
			if filename.find('*') < 0:
				filename = '.'.join(filename.split('.', 1).insert(1, '*'))
			for n in range(len(gimpDocument.layers)):
				saveLayer(gimpDocument, n, filename)
		else:
			saveLayer(gimpDocument, int(layer), filename)
	if args.save:
		gimpDocument.save(args.save)


def showLayer(image, l):
	""" show a layer """
	if image is None:
		print('No image for layer', l)
	else:
		print('Showing layer', l)
		image.show()


def saveLayer(gimpDoc, l, fileName):
	""" save a layer """
	iteration = gimpDoc.layers[l].image
	if iteration is None:
		print('No image for layer', l)
	else:
		fn2 = fileName.replace('*', str(l))
		print('saving layer', fn2)
		iteration.save(fn2)



#!/usr/bin/env python3
"""
Pure python implementation of the gimp vbr brush format
"""

import argparse

class GimpVbrBrush:
	"""
	Pure python implementation of the gimp vbr brush format

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/vbr.txt
	"""

	BRUSH_SHAPES = ["circle", "square", "diamond"]

	def __init__(self, filename=None):
		self.version = 1.0
		self.name = ''
		self.spacing = 0
		self.radius = 50
		self.hardness = 1
		self.aspectRatio = 1
		self.angle = 0
		self.brushShape = None # one of the strings in self.BRUSH_SHAPES
		self.spikes = None
		if filename is not None:
			self.load(filename)

	def load(self, filename):
		"""
		load a gimp file

		:param filename: can be a file name or a file-like object
		"""
		if hasattr(filename, 'read'):
			self.filename = filename.name
			f = filename
		else:
			self.filename = filename
			f = open(filename, 'r')
		data = f.read()
		f.close()
		self._decode_(data)

	@property
	def image(self):
		"""
		this parametric brush converted to a useable PIL image
		"""
		raise NotImplementedError() # TODO:

	def _decode_(self, data):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		"""
		data = [s.strip() for s in data.split('\n')]
		if data[0] != "GIMP-VBR":
			raise Exception('File format error.  Magic value mismatch.')
		self.version = float(data[1])
		if self.version == 1.0:
			self.name = data[2] # max len 255 bytes
			self.spacing = float(data[3])
			self.radius = float(data[4])
			self.hardness = float(data[5])
			self.aspectRatio = float(data[6])
			self.angle = float(data[7])
		elif self.version == 1.5:
			self.name = data[2] # max len 255 bytes
			self.brushShape = data[3] # one of the strings in self.BRUSH_SHAPES
			self.spacing = float(data[4])
			self.radius = float(data[5])
			self.spikes = float(data[6])
			self.hardness = float(data[7])
			self.aspectRatio = float(data[8])
			self.angle = float(data[9])
		else:
			raise Exception('Unknown version ' + str(self.version))

	def toBytes(self):
		"""
		encode to a raw data stream
		"""
		data = []
		data.append("GIMP-VBR")
		data.append(str(self.version))
		if self.version == 1.0:
			data.append(str(self.name))
			data.append(str(self.spacing))
			data.append(str(self.radius))
			data.append(str(self.hardness))
			data.append(str(self.aspectRatio))
			data.append(str(self.angle))
		elif self.version == 1.5:
			data.append(str(self.name))
			data.append(str(self.brushShape))
			data.append(str(self.spacing))
			data.append(str(self.radius))
			data.append(str(self.spikes))
			data.append(str(self.hardness))
			data.append(str(self.aspectRatio))
			data.append(str(self.angle))
		return ('\n'.join(data) + '\n').encode('utf-8')

	def save(self, toFilename=None, toExtension=None):
		"""
		save this gimp image to a file
		"""
		asImage = False
		if toExtension is None:
			if toFilename is not None:
				toExtension = toFilename.rsplit('.', 1)
				if len(toExtension) > 1:
					toExtension = toExtension[-1]
				else:
					toExtension = None
		if toExtension is not None and toExtension != 'vbr':
			asImage = True
		if asImage:
			self.image.save(toFilename)
		else:
			if not hasattr(toFilename, 'write'):
				f = open(toFilename, 'wb')
			f.write(self.toBytes())

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		if self.filename is not None:
			ret.append('Filename: ' + self.filename)
		ret.append('Name: ' + str(self.name))
		ret.append('Version: ' + str(self.version))
		ret.append('Spacing: ' + str(self.spacing))
		ret.append('Radius: ' + str(self.radius))
		ret.append('Hardness: ' + str(self.hardness))
		ret.append('Aspect ratio: ' + str(self.aspectRatio))
		ret.append('Angle: ' + str(self.angle))
		ret.append('Brush Shape: ' + str(self.brushShape))
		ret.append('Spikes: ' + str(self.spikes))
		return '\n'.join(ret)

	def __eq__(self, other):
		"""
		perform a comparison
		"""
		if other.name != self.name:
			return False
		if other.version != self.version:
			return False
		if other.spacing != self.spacing:
			return False
		if other.radius != self.radius:
			return False
		if other.hardness != self.hardness:
			return False
		if other.aspectRatio != self.aspectRatio:
			return False
		if other.angle != self.angle:
			return False
		if other.brushShape != self.brushShape:
			return False
		if other.spikes != self.spikes:
			return False
		return True


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("gimpVbrBrush.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	parser.add_argument("--dump", action="store_true",
	help="dump info about this file")
	args = parser.parse_args()

	gimpVbrBrush = GimpVbrBrush(args.xcfdocument)

	if args.dump:
		print(gimpVbrBrush)



#!/usr/bin/env python3
"""
Gimp Image Pipe Format

The gih format is use to store a series of brushes, and some extra info
for how to use them.
"""
import argparse
from .binaryIO import IO
from .gimpGbrBrush import GimpGbrBrush


class GimpGihBrushSet:
	"""
	Gimp Image Pipe Format

	The gih format is use to store a series of brushes, and some extra info
	for how to use them.

	See:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/gih.txt
	"""
	def __init__(self, filename=None):
		self.filename = None
		self.name = ''
		self.params = {}
		self.brushes = []
		if filename is not None:
			self.load(filename)

	def load(self, filename):
		"""
		load a gimp file

		:param filename: can be a file name or a file-like object
		"""
		if hasattr(filename, 'read'):
			self.filename = filename.name
			f = filename
		else:
			self.filename = filename
			f = open(filename, 'rb')
		data = f.read()
		f.close()
		self._decode_(data)

	def _decode_(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		self.name = io.textLine
		secondLine = io.textLine.split(' ')
		self.params = {}
		numBrushes = int(secondLine[0])
		# everything that's left is a gimp-image-pipe-parameters parasite
		for i in range(1, len(secondLine)):
			param = secondLine[i].split(':', 1)
			self.params[param[0].strip()] = param[1].strip()
		self.brushes = []
		for _ in range(numBrushes):
			b = GimpGbrBrush()
			io.index = b._decode_(
			io.data,
			io.index) # TODO: broken.  For some reson there is extra data between brushes!
			self.brushes.append(b)
		return io.index

	def toBytes(self):
		"""
		encode this object to a byte array
		"""
		io = IO()
		io.textLine = self.name
		# add the second line of data
		secondLine = [str(len(self.brushes))]
		for key in self.params:
			secondLine.append(key + ':' + str(self.params[key]))
		io.textLine = ' '.join(secondLine)
		# add the brushes
		for brush in self.brushes:
			io.addBytes(brush.toBytes())
		return io.data

	def save(self, toFilename=None):
		"""
		save this gimp image to a file
		"""
		if not hasattr(toFilename, 'write'):
			f = open(toFilename, 'wb')
		f.write(self.toBytes())

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		if self.filename is not None:
			ret.append('Filename: ' + self.filename)
		ret.append('Name: ' + str(self.name))
		for k, v in list(self.params.items()):
			ret.append(k + ': ' + str(v))
		for i in range(len(self.brushes)):
			ret.append('Brush ' + str(i))
			ret.append(self.brushes[i].__repr__(indent + '\t'))
		return ('\n' + indent).join(ret)


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("gimpGihBrushSet.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--dump", action="store_true",
	help="dump info about this file")
	group.add_argument("--show", action="store",
	help="show the brush image (n or * for all layers)")
	group.add_argument("--save", action="store",
	help="save out the brush set (n,out-file.png)")
	args = parser.parse_args()

	gimpGihBrushSet = GimpGihBrushSet(args.xcfdocument)

	if args.dump:
		print(gimpGihBrushSet)
	if args.show:
		if args.show == '*':
			for iterationM in range(len(gimpGihBrushSet.brushes)):
				gimpGihBrushSet.brushes[iterationM].image.show()
		else:
			gimpGihBrushSet.brushes[int(args.show)].image.show()
	if args.save:
		indexM, filenameM = args.save.split(',', 1)
		if filenameM.find('*') < 0:
			filenameM = '*.'.join(filenameM.split('.', 1))
		if indexM == '*':
			for iterationM in range(len(gimpGihBrushSet.brushes)):
				fn2 = filenameM.replace('*', str(iterationM))
				gimpGihBrushSet.brushes[iterationM].image.save(fn2)
		else:
			fn2 = filenameM.replace('*', iterationM)
			gimpGihBrushSet.brushes[int(indexM)].image.save(fn2)



#!/usr/bin/env python3
"""
Pure python implementation of the gimp file formats
"""
import argparse
from .gimpXcfDocument import GimpDocument
register = False

class GimpFormatPlugin:
	"""
	Pure python implementation of the gimp file formats
	"""

# ========= automatically add format info for pyformatgenie =========
if register:
	try:
		# will run this every time this module is loaded
		import pyformatgenie
		pfg = pyformatgenie.PyFormatGenie()
		pfg.add(GimpFormatPlugin)
		# and the most important part...
		pfg.save()
	except ImportError as e:
		# pyformatgenie is not installed (yet?). Continue with whatever you came here for.
		pass


if __name__ == '__main__':
	""" CLI Entry Point """
	parser = argparse.ArgumentParser("gimpFormat.py")
	parser.add_argument("xcfdocument", action="store",
	help="xcf file to act on")
	group = parser.add_mutually_exclusive_group()
	group.add_argument("--dump", action="store_true",
	help="dump info about this file")
	group.add_argument("--showLayer", action="store",
	help="show layer(s) (use * for all)")
	group.add_argument("--saveLayer", action="store",
	help="n,out.jpg save layer(s) out to file")
	args = parser.parse_args()

	gimpDocument = GimpDocument(args.xcfdocument)

	if args.dump:
		print(gimpDocument)
	if args.showLayer:
		if args.showLayer == '*':
			for layer in range(len(gimpDocument.layers)):
				iterationM = gimpDocument.layers[layer].image
				showLayer(iterationM, layer)
		else:
			iterationM = gimpDocument.layers[int(args.showLayer)].image
			showLayer(iterationM, int(args.showLayer))
	if args.saveLayer:
		layer = args.saveLayer.split(',', 1)
		if len(layer) > 1:
			filenameM = layer[1]
		else:
			filenameM = 'layer *.png'
		layer = args.saveLayer[0]
		if layer == '*':
			if filenameM.find('*') < 0:
				filenameM = '.'.join(filenameM.split('.', 1).insert(1, '*'))
			for layer in range(len(gimpDocument.layers)):
				saveLayer(gimpDocument, layer, filenameM)
		else:
			saveLayer(gimpDocument, layer, filenameM)


def showLayer(iteration, l):
	""" show a layer """
	if iteration is None:
		print('No image for layer', l)
	else:
		print('Showing layer', l)
		iteration.show()


def saveLayer(gimpDoc, l, filename):
	""" save a layer """
	iteration = gimpDoc.layers[l].image
	if iteration is None:
		print('No image for layer', l)
	else:
		fn2 = filename.replace('*', str(l))
		print('saving layer', fn2)
		iteration.save(fn2)



""" gimpformats

Forked from https://github.com/TheHeadlessSourceMan/gimpFormats

A pure python implementation of the GIMP XCF image format.

Use this to interact with GIMP image formats
"""
from .gimpFormat import *
from .gimpGbrBrush import *
from .gimpGgrGradient import *
from .gimpGihBrushSet import *
from .gimpGpbBrush import *
from .gimpGtpToolPreset import *
from .gimpImageInternals import *
from .gimpIOBase import *
from .gimpParasites import *
from .gimpPatPattern import *
from .gimpVbrBrush import *
from .gimpVectors import *
from .gimpXcfDocument import *
from .gimpGplPalette import *



#!/usr/bin/env python3
"""
Parasites are arbitrary (meta)data strings that can be attached to a document tree item

They are used to store things like last-used plugin settings, gamma adjuetments, etc.

Format of known parasites:
	https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/parasites.txt
"""
from .binaryIO import IO

#TODO: how to best use these for our puproses??
KNOWN_DOCUMENT_PARASITES = [
"jpeg-save-defaults", "png-save-defaults", "<plug-in>/_fu_data", "exif-orientation-rotate"]
KNOWN_IMAGE_PARASITES = [
"gimp-comment", "gimp-brush-name", "gimp-brush-pipe-name", "gimp-brush-pipe-parameters",
"gimp-image-grid", "gimp-pattern-name", "tiff-save-options", "jpeg-save-options",
"jpeg-exif-data", "jpeg-original-settings", "gamma", "chromaticity", "rendering-intent",
"hot-spot", "exif-data", "gimp-metadata", "icc-profile", "icc-profile-name", "decompose-data",
"print-settings", "print-page-setup", "dcm/XXXX-XXXX-AA"]
KNOWN_LAYER_PARASITES = ["gimp-text-layer", "gfig"]


class GimpParasite:
	"""
	Parasites are arbitrary (meta)data strings that can be attached to a document tree item

	They are used to store things like last-used plugin settings, gamma adjuetments, etc.

	Format of known parasites:
		https://gitlab.gnome.org/GNOME/gimp/blob/master/devel-docs/parasites.txt
	"""
	def __init__(self):
		self.name = ''
		self.flags = 0
		self.data = None

	def fromBytes(self, data, index=0):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO(data, index)
		self.name = io.sz754
		self.flags = io.u32
		dataLength = io.u32
		self.data = io.getBytes(dataLength)
		return io.index

	def toBytes(self):
		"""
		decode a byte buffer

		:param data: data buffer to decode
		:param index: index within the buffer to start at
		"""
		io = IO()
		io.sz754 = self.name
		io.u32 = self.flags
		io.u32 = len(self.data)
		io.addBytes(self.data)
		return io.data

	def __repr__(self, indent=''):
		"""
		Get a textual representation of this object
		"""
		ret = []
		ret.append('Name: ' + str(self.name))
		ret.append('Flags: ' + str(self.flags))
		ret.append('Data Len: ' + str(len(self.data)))
		return indent + (('\n' + indent).join(ret))



#!/usr/bin/env python3
"""
Base binary I/O helper.

Does boilerplate things like reading the next uint32 from the document
"""
import struct


class IO:
	"""
	Class to handle i/o to a byte buffer or file-like object
	"""
	def __init__(self, data=None, idx=0, littleEndian=False, boolSize=8, stringEncoding='U'):
		"""
		:param data: can be a data buffer or a file-like object
		:param idx: start reading/writing the data at the given index
		:param littleEndian: whether the default is big-endian or little-endian
		:param boolSize: how many default bits to use for a bool (8,16,32,or 64)
		:param stringEncoding: default string encoding A=Ascii, U=UTF-8, W-Unicode wide
		"""
		if data is None:
			data = bytearray()
		self.data = data
		self.index = idx
		self._contexts = []
		self.littleEndian = littleEndian
		self.boolSize = boolSize
		self.stringEncoding = stringEncoding # A=Ascii, U=UTF-8, W-Unicode wide

	@property
	def data(self):
		""" return data """
		return self._data

	@data.setter
	def data(self, data):
		""" set data """
		if not hasattr(data, "__getitem__"):
			raise Exception('ERR: incorrect type for data buffer' + str(type(data)))
		self._data = data

	def beginContext(self, newIndex):
		"""
		Start a new context where the index can be changed all you want,
		and when endContext() is called, it will be restored to the current position
		"""
		self._contexts.append(newIndex)

	def endContext(self):
		"""
		Restore the index to the previous location where it was when beginContext() was called
		"""
		self.index = self._contexts.pop()

	def _write(self, size, fmt, data):
		"""
		general formatted write
		"""
		if self.index + size >= len(self.data):
			self.data.extend(bytearray((self.index + size) - len(self.data)))
		try:
			struct.pack_into(fmt, self.data, self.index, data)
			self.index += size
		except struct.error as e:
			raise Exception(type(data), fmt, size, data)

	@property
	def bool(self):
		""" return bool """
		if self.boolSize == 8:
			return self.bool8
		if self.boolSize == 16:
			return self.bool16
		if self.boolSize == 32:
			return self.bool32
		if self.boolSize == 64:
			return self.bool64
		raise Exception("Unknown bool size " + str(self.boolSize))

	@bool.setter
	def bool(self, ioBool):
		""" set bool """
		if self.boolSize == 8:
			self.bool8 = ioBool
		elif self.boolSize == 16:
			self.bool16 = ioBool
		elif self.boolSize == 32:
			self.bool32 = ioBool
		elif self.boolSize == 64:
			self.bool64 = ioBool
		else:
			raise Exception("Unknown bool size " + str(self.boolSize))

	@property
	def bool8(self):
		""" get bool8 """
		return self.u8 != 0

	@bool8.setter
	def bool8(self, ioBool):
		""" set a bool8 """
		self.u8 = ioBool

	@property
	def bool16(self):
		""" get bool16 """
		return self.u16 != 0

	@bool16.setter
	def bool16(self, ioBool):
		""" set bool16 """
		self.u16 = ioBool

	@property
	def bool32(self):
		""" get bool32 """
		return self.u32 != 0

	@bool32.setter
	def bool32(self, ioBool):
		""" set bool32 """
		self.u32 = ioBool

	@property
	def bool64(self):
		""" get bool64 """
		return self.u64 != 0

	@bool64.setter
	def bool64(self, ioBool):
		""" set bool64 """
		self.u64 = ioBool

	@property
	def byte(self):
		""" get byte """
		return self.i8

	@byte.setter
	def byte(self, byte):
		""" set byte """
		self.i8 = byte

	@property
	def unsignedByte(self):
		""" get unsigned byte """
		return self.u8

	@unsignedByte.setter
	def unsignedByte(self, byte):
		""" set unsigned byte """
		self.u8 = byte

	@property
	def word(self):
		""" get a word """
		return self.i16

	@word.setter
	def word(self, word):
		""" set a word """
		self.i16 = word

	@property
	def unsignedWord(self):
		""" get an unsigned word """
		return self.u16

	@unsignedWord.setter
	def unsignedWord(self, unsignedWord):
		""" set an unsigned word """
		self.u16 = unsignedWord

	@property
	def dword(self):
		""" get a dword """
		return self.i32

	@dword.setter
	def dword(self, dword):
		""" set a dword """
		self.i32 = dword

	@property
	def unsignedDword(self):
		""" get a unsigned dword """
		return self.u32

	@unsignedDword.setter
	def unsignedDword(self, unsignedDword):
		""" set an unsigned dword """
		self.u32 = unsignedDword

	@property
	def qword(self):
		""" get a qword """
		return self.i64

	@qword.setter
	def qword(self, qword):
		self.i64 = qword

	@property
	def unsignedQword(self):
		""" set an unsigned qword """
		return self.u64

	@unsignedQword.setter
	def unsignedQword(self, unsignedQword):
		self.u64 = unsignedQword

	@property
	def i8(self):
		""" get an int8 """
		if self.littleEndian:
			return self.i8le
		return self.i8be

	@i8.setter
	def i8(self, i8):
		""" set an int8 """
		if self.littleEndian:
			self.i8le = i8
		else:
			self.i8be = i8

	@property
	def u8(self):
		""" get an unsigned int """
		if self.littleEndian:
			return self.u8le
		return self.u8be

	@u8.setter
	def u8(self, u8):
		""" set an unsigned int """
		if self.littleEndian:
			self.u8le = u8
		else:
			self.u8be = u8

	@property
	def i16(self):
		""" get an int16 """
		if self.littleEndian:
			return self.i16le
		return self.i16be

	@i16.setter
	def i16(self, i16):
		""" set an int16 """
		if self.littleEndian:
			self.i16le = i16
		else:
			self.i16be = i16

	@property
	def u16(self):
		""" get an uint16 """
		if self.littleEndian:
			return self.i16le
		return self.i16be

	@u16.setter
	def u16(self, u16):
		""" set an unint16 """
		if self.littleEndian:
			self.u16le = u16
		else:
			self.u16be = u16

	@property
	def i32(self):
		""" get an int32 """
		if self.littleEndian:
			return self.i32le
		return self.i32be

	@i32.setter
	def i32(self, i32):
		"""set an int32 """
		if self.littleEndian:
			self.i32le = i32
		else:
			self.i32be = i32

	@property
	def u32(self):
		""" get a uint32 """
		if self.littleEndian:
			return self.i32le
		return self.i32be

	@u32.setter
	def u32(self, u32):
		""" set a unint32 """
		if self.littleEndian:
			self.u32le = u32
		else:
			self.u32be = u32

	@property
	def i64(self):
		""" get an int64 """
		if self.littleEndian:
			return self.i64le
		return self.i64be

	@i64.setter
	def i64(self, i64):
		""" set an int64 """
		if self.littleEndian:
			self.i64le = i64
		else:
			self.i64be = i64

	@property
	def u64(self):
		""" get a uint64 """
		if self.littleEndian:
			return self.i64le
		return self.i64be

	@u64.setter
	def u64(self, u64):
		""" set a uint64 """
		if self.littleEndian:
			self.u64le = u64
		else:
			self.u64be = u64

	@property
	def float32(self):
		""" get a float32 """
		if self.littleEndian:
			return self.float32le
		return self.float32be

	@float32.setter
	def float32(self, float32):
		""" set a float32 """
		if self.littleEndian:
			self.float32le = float32
		else:
			self.float32be = float32

	@property
	def float64(self):
		""" get a float64 """
		if self.littleEndian:
			return self.float64le
		return self.float64be

	@float64.setter
	def float64(self, float64):
		""" set a float64 """
		if self.littleEndian:
			self.float64le = float64
		else:
			self.float64be = float64

	@property
	def u8be(self):
		"""
		read the next uint8 and advance the index
		"""
		d = struct.unpack('>B', self.data[self.index:self.index + 1])[0]
		self.index += 1
		return d

	@u8be.setter
	def u8be(self, u8be):
		""" set the uint8 """
		self._write(1, '>B', u8be)

	@property
	def u8le(self):
		"""
		read the next uint8 and advance the index
		"""
		d = struct.unpack('<B', self.data[self.index:self.index + 1])[0]
		self.index += 1
		return d

	@u8le.setter
	def u8le(self, u8le):
		""" set the uint8 """
		self._write(1, '<B', u8le)

	@property
	def i8le(self):
		"""
		read the next signed int8 and advance the index
		"""
		d = struct.unpack('<b', self.data[self.index:self.index + 1])[0]
		self.index += 1
		return d

	@i8le.setter
	def i8le(self, i8le):
		""" set the int8 """
		self._write(1, '<b', i8le)

	@property
	def i8be(self):
		"""
		read the next signed int8 and advance the index
		"""
		d = struct.unpack('>b', self.data[self.index:self.index + 1])[0]
		self.index += 1
		return d

	@i8be.setter
	def i8be(self, i8be):
		""" set the int8 """
		self._write(1, '>b', i8be)

	@property
	def u16be(self):
		"""
		read the next uint16 and advance the index
		"""
		d = struct.unpack('>H', self.data[self.index:self.index + 2])[0]
		self.index += 2
		return d

	@u16be.setter
	def u16be(self, u16be):
		""" set the uint16 """
		self._write(2, '>H', u16be)

	@property
	def u16le(self):
		"""
		read the next uint16 and advance the index
		"""
		d = struct.unpack('<H', self.data[self.index:self.index + 2])[0]
		self.index += 2
		return d

	@u16le.setter
	def u16le(self, u16le):
		""" set the uint16 """
		self._write(2, '<H', u16le)

	@property
	def i16le(self):
		"""
		read the next signed int16 and advance the index
		"""
		d = struct.unpack('<h', self.data[self.index:self.index + 2])[0]
		self.index += 2
		return d

	@i16le.setter
	def i16le(self, i16le):
		""" set the int16 """
		self._write(2, '<h', i16le)

	@property
	def i16be(self):
		"""
		read the next signed int16 and advance the index
		"""
		d = struct.unpack('>h', self.data[self.index:self.index + 2])[0]
		self.index += 2
		return d

	@i16be.setter
	def i16be(self, i16be):
		""" set the int16 """
		self._write(2, '>h', i16be)

	@property
	def u32be(self):
		"""
		read the next uint32 and advance the index
		"""
		d = struct.unpack('>I', self.data[self.index:self.index + 4])[0]
		self.index += 4
		return d

	@u32be.setter
	def u32be(self, u32be):
		""" set the uint32 """
		self._write(4, '>I', int(u32be))

	@property
	def u32le(self):
		"""
		read the next uint32 and advance the index
		"""
		d = struct.unpack('<I', self.data[self.index:self.index + 4])[0]
		self.index += 4
		return d

	@u32le.setter
	def u32le(self, u32le):
		""" set the uint32 """
		self._write(4, '<I', u32le)

	@property
	def i32le(self):
		"""
		read the next signed int32 and advance the index
		"""
		d = struct.unpack('<i', self.data[self.index:self.index + 4])[0]
		self.index += 4
		return d

	@i32le.setter
	def i32le(self, i32le):
		""" set the int32 """
		self._write(4, '<i', i32le)

	@property
	def i32be(self):
		"""
		read the next signed int32 and advance the index
		"""
		try:
			d = struct.unpack('>i', self.data[self.index:self.index + 4])[0]
		except Exception as _e:
			raise Exception(
			str(self.index) + ' ' + str(len(self.data)) + ' ' +
			str(self.data[self.index:self.index + 4]))
		self.index += 4
		return d

	@i32be.setter
	def i32be(self, i32be):
		""" set the int32 """
		self._write(4, '>i', int(i32be))

	@property
	def u64be(self):
		"""
		read the next uint64 and advance the index
		"""
		d = struct.unpack('>Q', self.data[self.index:self.index + 8])[0]
		self.index += 8
		return d

	@u64be.setter
	def u64be(self, u64be):
		""" set the uint64 """
		self._write(8, '>Q', u64be)

	@property
	def u64le(self):
		"""
		read the next uint64 and advance the index
		"""
		d = struct.unpack('<Q', self.data[self.index:self.index + 8])[0]
		self.index += 8
		return d

	@u64le.setter
	def u64le(self, u64le):
		""" set the uint64 """
		self._write(8, '<Q', u64le)

	@property
	def i64le(self):
		"""
		read the next signed int64 and advance the index
		"""
		d = struct.unpack('<q', self.data[self.index:self.index + 8])[0]
		self.index += 8
		return d

	@i64le.setter
	def i64le(self, i64le):
		""" set the int64 """
		self._write(8, '<q', i64le)
		self.index += 8

	@property
	def i64be(self):
		"""
		read the next signed int64 and advance the index
		"""
		d = struct.unpack('>q', self.data[self.index:self.index + 8])[0]
		self.index += 8
		return d

	@i64be.setter
	def i64be(self, i64be):
		""" set the int64 """
		self._write(8, '>q', i64be)

	@property
	def float(self):
		""" get a float """
		return self.float32

	@float.setter
	def float(self, f):
		""" set a float """
		self.float32 = f

	@property
	def double(self):
		""" get a double """
		return self.float64

	@double.setter
	def double(self, f):
		""" set a double """
		self.float64 = f

	@property
	def float32be(self):
		"""
		read the next 32 bit float and advance the index
		"""
		d = struct.unpack('>f', self.data[self.index:self.index + 4])[0]
		self.index += 4
		return d

	@float32be.setter
	def float32be(self, float32be):
		""" set a 32 bit float """
		self._write(4, '>f', float32be)

	@property
	def float32le(self):
		"""
		read the next 32 bit float and advance the index
		"""
		d = struct.unpack('<f', self.data[self.index:self.index + 4])[0]
		self.index += 4
		return d

	@float32le.setter
	def float32le(self, float32le):
		""" set a 32 bit float """
		self._write(4, '<f', float32le)

	@property
	def float64be(self):
		"""
		read the next 64 bit float and advance the index
		"""
		d = struct.unpack('>d', self.data[self.index:self.index + 8])[0]
		self.index += 8
		return d

	@float64be.setter
	def float64be(self, float64be):
		""" set a 64 bit float """
		self._write(8, '>d', float64be)

	@property
	def float64le(self):
		"""
		read the next 64 bit float and advance the index
		"""
		d = struct.unpack('<d', self.data[self.index:self.index + 8])[0]
		self.index += 8
		return d

	@float64le.setter
	def float64le(self, float64le):
		""" set a 64 bit float """
		self._write(8, '<d', float64le)

	def getBytes(self, nbytes):
		"""
		grab some raw bytes and advance the index
		"""
		d = self.data[self.index:self.index + nbytes]
		self.index += nbytes
		return d

	def addBytes(self, ioBytes):
		"""
		add some raw bytes and advance the index

		alias for setBytes()

		:param bytes: can be a string, bytearray, or another IO object
		"""
		self.setBytes(ioBytes)

	def setBytes(self, ioBytes):
		"""
		add some raw bytes and advance the index

		alias for addBytes()

		:param ioBytes: can be a string, bytearray, or another IO object
		"""
		if isinstance(ioBytes, IO):
			ioBytes = ioBytes.data
		if isinstance(ioBytes, str):
			ioBytes = bytearray(ioBytes, encoding="utf-8")
		if self.index >= len(self.data):
			# if we're at the end, simply extend the buffer
			self.data.extend(ioBytes)
			self.index += len(ioBytes)
		else:
			if self.index + len(ioBytes) >= len(self.data):
				self.data.extend(bytearray((self.index + len(ioBytes)) - len(self.data)))
			for b in ioBytes:
				self.data[self.index] = b
				self.index += 1

	def _sz754(self, encoding):
		"""
		Read the next string conforming to IEEE 754 and advance the index

		Note, string format is:
			uint32   n+1  Number of bytes that follow, including the zero byte
			byte[n]  ...  String data in Unicode, encoded using UTF-8
			byte     0    Zero marks the end of the string.
		or simply uint32=0 for empty string
		"""
		nchars = self.u32
		if nchars == 0:
			return ''
		d = self.data[self.index:self.index + nchars - 1]
		self.index += nchars
		if encoding == 'A':
			return d
		if encoding == 'U':
			return d.decode('UTF-8', errors='replace')
		if encoding == 'W':
			return d.decode('UTF-16', errors='replace')
		raise Exception()

	def _sz754set(self, sz754, _encoding):
		'''_sz754set'''
		self.u32 = len(sz754)
		self.setBytes(sz754)
		self.u8 = 0

	@property
	def sz754(self):
		'''sz754'''
		return self._sz754(self.stringEncoding)

	@sz754.setter
	def sz754(self, sz754):
		'''set sz754'''
		return self._sz754set(sz754, self.stringEncoding)

	@property
	def sz754A(self):
		'''sz754A'''
		return self._sz754('A')

	@sz754A.setter
	def sz754A(self, sz754):
		'''set sz754A'''
		return self._sz754set(sz754, 'A')

	@property
	def sz754W(self):
		'''sz754W'''
		return self._sz754('W')

	@sz754W.setter
	def sz754W(self, sz754):
		'''set sz754W'''
		return self._sz754set(sz754, 'W')

	@property
	def sz754U(self):
		'''sz754U'''
		return self._sz754('U')

	@sz754U.setter
	def sz754U(self, sz754):
		'''set sz754U'''
		return self._sz754set(sz754, 'U')

	def _readUntil(self, until, encoding='A'):
		"""
		:param until: must be within the ascii character set
		:param encoding: one of A (ascii), U (UTF-8) or W (UCS-2)
		"""
		d = []
		if encoding == 'A':
			encoding = 'ascii'
		elif encoding == 'U':
			encoding = 'UTF-8'
		elif encoding == 'W':
			encoding = 'UCS-2'
		else:
			raise Exception('bogus encoding')
		until = until.encode('ascii')[0] # always ascii
		while True:
			c = self.data[self.index]
			self.index += 1
			if encoding == 'W': # get one more byte
				d.append(c)
				c = self.data[self.index]
				self.index += 1
			if c == until:
				break
			d.append(c)
		return bytes(d).decode(encoding, errors='replace')

	@property
	def textLine(self):
		'''textLine'''
		ret = self._readUntil('\n', self.stringEncoding)
		if ret[-1] == '\r':
			ret = ret[-1]
		return ret

	@textLine.setter
	def textLine(self, text):
		self.setBytes(text)
		if text[-1] != '\n':
			self.setBytes('\n')

	@property
	def textLineA(self):
		'''textLineA'''
		ret = self._readUntil('\n', 'A')
		if ret[-1] == '\r':
			ret = ret[-1]
		return ret

	@textLineA.setter
	def textLineA(self, text):
		self.setBytes(text)
		if text[-1] != '\n':
			self.setBytes('\n')

	@property
	def textLineW(self):
		'''textLineW'''
		ret = self._readUntil('\n', 'W')
		if ret[-1] == '\r':
			ret = ret[-1]
		return ret

	@textLineW.setter
	def textLineW(self, text):
		self.setBytes(text)
		if text[-1] != '\n':
			self.setBytes('\0\n')

	@property
	def textLineU(self):
		'''textLineU'''
		ret = self._readUntil('\n', 'U')
		if ret[-1] == '\r':
			ret = ret[-1]
		return ret

	@textLineU.setter
	def textLineU(self, text):
		self.setBytes(text)
		if text[-1] != '\n':
			self.setBytes('\n')

	@property
	def cString(self):
		'''cString'''
		return self._readUntil('\0', self.stringEncoding)

	@cString.setter
	def cString(self, text):
		self.setBytes(text)
		self.setBytes('\0')

	@property
	def cStringA(self):
		'''cStringA'''
		return self._readUntil('\0', 'A')

	@cString.setter
	def cStringA(self, text):
		self.setBytes(text)
		self.setBytes('\0')

	@property
	def cStringW(self):
		'''cStringW'''
		return self._readUntil('\0', 'W')

	@cString.setter
	def cString(self, text):
		self.setBytes(text)
		self.setBytes('\0\0')

	@property
	def cStringU(self):
		'''cStringU'''
		return self._readUntil('\0', 'U')

	@cString.setter
	def cString(self, text):
		self.setBytes(text)
		self.setBytes('\0')
